#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Driver for the NSDE algorithm.

This driver uses the differential_evolution Python package to provide the logic of the main
differential evolution algorithms, developed by D. de Vries.
"""
import copy
import itertools
import numpy as np
import openmdao
import os

from openmdao.core.analysis_error import AnalysisError
from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.utils.mpi import MPI

try:
    from tqdm import tqdm
except ModuleNotFoundError:

    def tqdm(i, _):
        return i


from nsde import NSDE, EvolutionStrategy
from nsde.strategies import mutation_strategies, crossover_strategies, repair_strategies

if not MPI:
    rank = 0
else:
    rank = MPI.COMM_WORLD.rank


def progress_string(de):
    s = " "
    if tqdm is None:
        s += "gen: {:>5g} / {}, ".format(de.generation, de.max_gen)
    if de.n_obj == 1:
        s += (
            "f*: {:> 10.4g}, "
            "dx: {:> 10.4g}, "
            "df: {:> 10.4g}".format(
                de.fit[0, 0],
                de.dx,
                de.df[0] if isinstance(de.df, np.ndarray) else de.df
            )
        )
    else:
        s += "1 - hv: {:> 10.4g}".format(1 - de.hv / (1.1 ** de.n_obj))
    if de.n_con > 0:
        s += f", feasible: {np.count_nonzero(np.all(de.con <= 1e-6, axis=1).flatten()):>4d}/{de.n_pop}"
    return s.replace("\n", "")


class NSDEDriver(Driver):
    """
    Driver for the NSDE algorithm.

    Attributes
    ----------
    _concurrent_pop_size : int
        Number of points to run concurrently when model is a parallel one.
    _concurrent_color : int
        Color of current rank when running a parallel model.
    _de : NSDE
        Differential evolution algorithm.
    _desvar_idx : dict
        Keeps track of the indices for each desvar, since GeneticAlgorithm sees an array of
        design variables.
    _es : EvolutionStrategy
        Evolution strategy to use when evolving the population of the differential evolution algorithm.
    _seed : int
         Seed number which controls the seed and random draws.
    """

    def __init__(self, **kwargs):
        """
        Initialize the NSDE driver.

        Parameters
        ----------
        **kwargs : dict of keyword arguments
            Keyword arguments that will be mapped into the Driver options.
        """
        super(NSDEDriver, self).__init__(**kwargs)

        # What we support
        self.supports["inequality_constraints"] = True
        self.supports["equality_constraints"] = True
        self.supports["multiple_objectives"] = True

        # What we don't support yet
        self.supports["integer_design_vars"] = False
        self.supports["two_sided_constraints"] = False
        self.supports["linear_constraints"] = False
        self.supports["simultaneous_derivatives"] = False
        self.supports["active_set"] = False

        self._desvar_idx = {}
        self._es = None
        self._de = None

        # random state can be set for predictability during testing
        if "NSDEDriver_seed" in os.environ:
            self._seed = int(os.environ["NSDEDriver_seed"])
        else:
            self._seed = None

        # Support for Parallel models.
        self._concurrent_pop_size = 0
        self._concurrent_color = 0

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        self.options.declare(
            "strategy",
            default="rand-to-best/1/exp/random",
            values=[
                "/".join(strategy)
                for strategy in itertools.product(
                    list(mutation_strategies.keys()),
                    ["1", "2", "3"],
                    list(crossover_strategies.keys()),
                    list(repair_strategies.keys()),
                )
            ],
            desc="Evolution strategy to use for the differential evolution. "
            "An evolution strategy is made up of four parts in fixed order, separated by '/':"
            " mutation strategy ('rand', 'best', or 'rand-to-best'),"
            " number of individuals to involve in the mutation (1, 2, or 3),"
            " crossover strategy ('exp' or 'bin'), and"
            " repair strategy ('random' or 'clip'). "
            "A good introduction of these topics can be found here: "
            "https://pablormier.github.io/2017/09/05/a-tutorial-on-differential-evolution-with-python/",
        )
        self.options.declare(
            "Pm",
            desc="Mutation rate.",
            default=None,
            lower=0.0,
            upper=1.0,
            allow_none=True,
        )
        self.options.declare(
            "Pc",
            default=None,
            lower=0.0,
            upper=1.0,
            allow_none=True,
            desc="Crossover rate.",
        )
        self.options.declare(
            "adaptivity",
            default=2,
            values=[0, 1, 2],
            desc="Self-adaptivity setting:"
            " 0: mutation and crossover rates are fixed (no self-adaptivity);"
            " 1: mutation and crossover rates are optimized using Monte-Carlo approach; "
            " 2: mutation and crossover rates are optimized using evolutionary algorithm. ",
        )
        self.options.declare(
            "max_gen", default=1000, desc="Number of generations before termination."
        )
        self.options.declare(
            "tolx", default=1e-8, desc="Tolerance of the design vectors' spread."
        )
        self.options.declare(
            "tolf", default=1e-8, desc="Tolerance of the fitness spread."
        )
        self.options.declare(
            "tolc", default=1e-6, desc="Constraint violation tolerance."
        )
        self.options.declare(
            "pop_size",
            default=0,
            desc="Number of individuals (points) to use for the optimization. "
            "If set to 0, it will be calculated automatically as 5 x dimensionality.",
        )
        self.options.declare(
            "run_parallel",
            types=bool,
            default=False,
            desc="Set to True to execute the points in a generation in parallel.",
        )
        self.options.declare(
            "procs_per_model",
            default=1,
            lower=1,
            desc="Number of processors to give each model under MPI.",
        )
        self.options.declare(
            "show_progress",
            default=False,
            desc="Set to true if a progress bar should be shown.",
        )
        self.options.declare(
            "generation_callback",
            default=None,
            allow_none=True,
            desc="Callback which will be called for each generation."
            "Callable should have a single argument, which will "
            "be an instance of the NSDE class.",
        )
        self.options.declare(
            "initial_population",
            default=None,
            allow_none=True,
            desc="Initial population with which to start the optimization."
                 "This should be a 2D array with design vectors as rows and"
                 "one row per individual.",
        )

    def _setup_driver(self, problem):
        """
        Prepare the driver for execution.

        This is the final thing to run during setup.

        Parameters
        ----------
        problem : <Problem>
            Pointer to the containing problem.
        """
        super(NSDEDriver, self)._setup_driver(problem)

        model_mpi = None
        comm = self._problem.comm
        if self._concurrent_pop_size > 0:
            model_mpi = (self._concurrent_pop_size, self._concurrent_color)
        elif not self.options["run_parallel"]:
            comm = None

        self._es = EvolutionStrategy(self.options["strategy"])
        self._de = NSDE(
            strategy=self._es,
            mut=self.options["Pm"],
            crossp=self.options["Pc"],
            adaptivity=self.options["adaptivity"],
            max_gen=self.options["max_gen"],
            tolx=self.options["tolx"],
            tolf=self.options["tolf"],
            tolc=self.options["tolc"],
            n_pop=self.options["pop_size"],
            seed=self._seed,
            comm=comm,
            model_mpi=model_mpi,
        )

    def _setup_comm(self, comm):
        """
        Perform any driver-specific setup of communicators for the model.

        Here, we generate the model communicators.

        Parameters
        ----------
        comm : MPI.Comm or <FakeComm> or None
            The communicator for the Problem.

        Returns
        -------
        MPI.Comm or <FakeComm> or None
            The communicator for the Problem model.
        """
        procs_per_model = self.options["procs_per_model"]
        if MPI and self.options["run_parallel"]:

            full_size = comm.size
            size = full_size // procs_per_model
            if full_size != size * procs_per_model:
                raise RuntimeError(
                    "The total number of processors is not evenly divisible by the "
                    "specified number of processors per model.\n Provide a "
                    "number of processors that is a multiple of %d, or "
                    "specify a number of processors per model that divides "
                    "into %d." % (procs_per_model, full_size)
                )
            color = comm.rank % size
            model_comm = comm.Split(color)

            # Everything we need to figure out which case to run.
            self._concurrent_pop_size = size
            self._concurrent_color = color

            return model_comm

        self._concurrent_pop_size = 0
        self._concurrent_color = 0
        return comm

    def _get_name(self):
        """
        Get name of current Driver.

        Returns
        -------
        str
            Name of current Driver.
        """
        return "NSDE"

    def get_de(self):
        """
        Get a copy of the driver's underlying DE class.

        Returns
        -------
        NSDE
            A copy of the driver's underlying DE class
        """
        return copy.copy(self._de)

    def run(self):
        """
        Execute the differential evolution algorithm.

        Returns
        -------
        boolean
            Failure flag; True if failed to converge, False is successful.
        """
        model = self._problem.model
        de = self._de

        de.strategy = EvolutionStrategy(self.options["strategy"])
        de.f = self.options["Pm"]
        de.cr = self.options["Pc"]
        de.adaptivity = self.options["adaptivity"]
        de.n_pop = self.options["pop_size"]
        de.max_gen = self.options["max_gen"]
        de.tolx = self.options["tolx"]
        de.tolf = self.options["tolf"]
        de.tolc = self.options["tolc"]

        self._check_for_missing_objective()

        # Size design variables.
        desvars = self._designvars
        desvar_vals = self.get_design_var_values()

        count = 0
        for name, meta in desvars.items():
            if name in self._designvars_discrete:
                val = desvar_vals[name]
                if np.isscalar(val):
                    size = 1
                else:
                    size = len(val)
            else:
                size = meta["size"]
            self._desvar_idx[name] = (count, count + size)
            count += size

        bounds = []
        x0 = np.empty(count)

        # Figure out bounds vectors and initial design vars
        for name, meta in desvars.items():
            i, j = self._desvar_idx[name]
            lb = meta["lower"]
            if isinstance(lb, float):
                lb = [lb] * (j - i)
            ub = meta["upper"]
            if isinstance(ub, float):
                ub = [ub] * (j - i)
            for k in range(j - i):
                bounds += [(lb[k], ub[k])]
            x0[i:j] = desvar_vals[name]

        de.init(self.objective_callback, bounds, self.options["initial_population"])
        if rank == 0 and self.options["show_progress"]:
            print(progress_string(de))
        if self.options["generation_callback"] is not None:
            self.options["generation_callback"](de)

        gen_iter = de
        if rank == 0 and self.options["show_progress"] and tqdm is not None:
            gen_iter = tqdm(gen_iter, total=self.options["max_gen"])

        for generation in gen_iter:
            if rank == 0 and self.options["show_progress"]:
                print(progress_string(generation))
                if self.options["generation_callback"] is not None:
                    self.options["generation_callback"](de)

        # Pull optimal parameters back into framework and re-run, so that
        # framework is left in the right final state
        best = de.pop[0]
        for name in desvars:
            i, j = self._desvar_idx[name]
            val = best[i:j]
            self.set_design_var(name, val)

        with RecordingDebugging(self._get_name(), self.iter_count, self) as rec:
            model.run_solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        return False

    def objective_callback(self, x):
        r"""
        Evaluate problem objective at the requested point.

        Parameters
        ----------
        x : ndarray
            Value of design variables.

        Returns
        -------
        f : ndarray
            Objective values
        g : ndarray
            Constraint values
        """
        model = self._problem.model

        for name in self._designvars:
            i, j = self._desvar_idx[name]
            self.set_design_var(name, x[i:j])

        # a very large number, but smaller than the result of nan_to_num in Numpy
        almost_inf = openmdao.INF_BOUND

        # Execute the model
        with RecordingDebugging(self._get_name(), self.iter_count, self) as rec:
            self.iter_count += 1
            try:
                model.run_solve_nonlinear()

            # Tell the optimizer that this is a bad point.
            except AnalysisError:
                model._clear_iprint()

            # Get the objective functions' values
            f = np.array(list(self.get_objective_values().values()))

            # Get the constraint violations
            g = np.array([])
            with np.errstate(
                divide="ignore"
            ):  # Ignore divide-by-zero warnings temporarily
                for name, val in self.get_constraint_values().items():
                    con = self._cons[name]
                    # The not used fields will either None or a very large number
                    # All constraints will be converted into standard <= 0 form.
                    if (con["lower"] is not None) and np.any(
                        con["lower"] > -almost_inf
                    ):
                        g = np.append(
                            g,
                            np.where(
                                con["lower"] == 0,
                                con["lower"] - val,
                                1 - val / con["lower"],
                            ).flatten(),
                        )
                    elif (con["upper"] is not None) and np.any(
                        con["upper"] < almost_inf
                    ):
                        g = np.append(
                            g,
                            np.where(
                                con["upper"] == 0,
                                val - con["upper"],
                                val / con["upper"] - 1,
                            ).flatten(),
                        )
                    elif (con["equals"] is not None) and np.any(
                        np.abs(con["equals"]) < almost_inf
                    ):
                        g = np.append(
                            g,
                            np.where(
                                con["equals"] == 0,
                                np.abs(con["equals"] - val),
                                np.abs(1 - val / con["equals"]),
                            ).flatten()
                            - self.options["tolc"],
                        )

            # Record after getting obj to assure they have
            # been gathered in MPI.
            rec.abs = 0.0
            rec.rel = 0.0

        return f.flatten(), g.flatten()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
__version__ = "0.0.11"

from .nsde import NSDE
from .strategies import EvolutionStrategy

try:
    from .openmdao import NSDEDriver
except ModuleNotFoundError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Definition of the NSDE algorithm"""
import numpy as np

try:
    from openmdao.utils.concurrent import concurrent_eval
except ModuleNotFoundError:
    import warnings

    warnings.warn("OpenMDAO is not installed. Concurrent evaluation is not available.")

from . import sorting, hv
from .strategies import EvolutionStrategy


def mpi_fobj_wrapper(fobj):
    """
    Wrapper for the objective function to keep track of individual indices when running under MPI.

    Parameters
    ----------
    fobj : callable
        Original objective function

    Returns
    -------
    callable
        Wrapped function which, in addition to x, takes the individual's index and returns it along with f
    """

    def wrapped(x, ii):
        return fobj(x), ii

    return wrapped


class NSDE:
    """
    Non-dominated Sorting Differential Evolution (NSDE) Algorithm.

    Attributes
    ----------
    fobj : callable
        Objective function.
        Should have a single argument of type array_like which corresponds to the design vector.
        Should have either a single float or 1D array output corresponding to the objective function value(s),
        or two array_like outputs, the first of which corresponds to the objective function value(s) and the second
        to the constraint violations.
        Constraints are assumed to be satisfied if constraint violations <= constraint tolerance.
    lb, ub : array_like
        Lower and upper bounds
    range : array_like
        Distances between the lower and upper bounds
    f, cr : float
        Mutation rate and crossover probabilities
    adaptivity : int
        Method of self-adaptivity.
            - 0: No self-adaptivity. Specified mutation rate and crossover probability are used.
            - 1: Simple self-adaptability. Mutation rate and crossover probability are optimized Mote-Carlo style.
            - 2: Complex self-adaptability. Mutation rate and crossover probability are mutated with specified strategy.
    max_gen : int
        Maximum number of generations
    tolx, tolf : float
        Tolerances on the design vectors' and objective function values' spreads
    tolc : float
        Constraint violation tolerance.
    n_dim : int
        Number of dimension of the problem
    n_pop : int
        Population size
    rng : np.random.Generator
        Random number generator
    comm : MPI communicator or None
        The MPI communicator that will be used objective evaluation for each generation
    model_mpi : None or tuple
        If the model in fobj is also parallel, then this will contain a tuple with the the
        total number of population points to evaluate concurrently, and the color of the point
        to evaluate on this rank
    strategy : EvolutionStrategy
        Evolution strategy to use for procreation
    pop : np.array
        List of the individuals' chromosomes making up the current population
    fit : np.array
        Fitness of the individuals in the population
    con : np.array
        Constraint violations of the individuals in the population
    generation : int
        Generation counter
    """

    def __init__(
        self,
        strategy=None,
        mut=0.85,
        crossp=1.0,
        adaptivity=0,
        max_gen=1000,
        tolx=1e-8,
        tolf=1e-8,
        tolc=1e-6,
        n_pop=None,
        seed=None,
        comm=None,
        model_mpi=None,
    ):
        self.fobj = None

        self.lb, self.ub = None, None
        self.range = 0

        self.f = mut
        self.cr = crossp

        self.max_gen = max_gen
        self.tolx = tolx
        self.tolf = tolf
        self.tolc = tolc

        self.n_dim = 0
        self.n_obj = 0
        self.n_con = 0
        self.n_pop = n_pop

        self.rng = np.random.default_rng(seed)

        if adaptivity not in [0, 1, 2]:
            raise ValueError("self_adaptivity must be one of (0, 1, 2).")
        self.adaptivity = adaptivity

        self.comm = comm
        self.model_mpi = model_mpi

        if strategy is None:
            self.strategy = EvolutionStrategy("rand-to-best/1/bin/random")
        elif isinstance(strategy, EvolutionStrategy):
            self.strategy = strategy
        elif isinstance(strategy, str):
            self.strategy = EvolutionStrategy(strategy)
        else:
            raise ValueError(
                "Argument `strategy` should be None, a str, or an instance of EvolutionStrategy."
            )

        self.pop = None
        self.fit = None
        self.con = None
        self.fronts = None
        self.dx, self.df, self.hv = np.inf, np.inf, np.inf

        self.pareto_lb = +np.inf
        self.pareto_ub = -np.inf

        self.generation = 0

        self._is_initialized = False
        self._running_under_mpi = comm is not None and hasattr(comm, "bcast")

    def init(self, fobj, bounds, pop=None):
        """
        Initialize the algorithm.

        Parameters
        ----------
        fobj : callable
            Objective function
        bounds : list of 2-tuples
            List of (lower, upper) bounds
        pop : None or array_like, optional
            Initial population. If None, it will be created at random.
        """
        # Set default values for the mutation and crossover parameters
        if self.f is None or 0.0 > self.f > 1.0:
            self.f = 0.85
        if self.cr is None or 0.0 > self.cr > 1.0:
            self.cr = 1.0

        # Prepare the objective function and compute the bounds and variable range
        self.fobj = fobj if self.comm is None else mpi_fobj_wrapper(fobj)
        self.lb, self.ub = np.asarray(bounds).T
        self.range = self.ub - self.lb

        # Compute the number of dimensions
        self.n_dim = len(bounds)

        def create_f_cr(adaptivity, f, cr, n, rng):
            # Create random mutation/crossover parameters if self-adaptivity is used
            if adaptivity == 0:
                f = f * np.ones(n)
                cr = cr * np.ones(n)
            elif adaptivity == 1:
                f = rng.uniform(size=n) * 0.9 + 0.1
                cr = rng.uniform(size=n)
            elif adaptivity == 2:
                f = rng.uniform(size=n) * 0.15 + 0.5
                cr = rng.uniform(size=n) * 0.15 + 0.5
            return f, cr

        adjust_pop = False
        if pop is not None:
            self.n_pop = pop.shape[0]
            self.pop = pop
            self.f, self.cr = create_f_cr(
                self.adaptivity, self.f, self.cr, self.n_pop, self.rng
            )
        else:
            if self.n_pop is None or self.n_pop <= 0:
                self.pop = self.rng.uniform(self.lb, self.ub, size=(1, self.n_dim))
                adjust_pop = True
                self.n_pop = 1
            else:
                self.pop = self.rng.uniform(
                    self.lb, self.ub, size=(self.n_pop, self.n_dim)
                )
                self.f, self.cr = create_f_cr(
                    self.adaptivity, self.f, self.cr, self.n_pop, self.rng
                )

        # Ensure all processors have the same population and mutation/crossover parameters
        if self._running_under_mpi:
            self.pop, self.f, self.cr = self.comm.bcast(
                (self.pop, self.f, self.cr), root=0
            )

        self.fit, self.con = self(self.pop)

        self.n_obj = self.fit.shape[1]
        if self.con is not None:
            self.n_con = self.con.shape[1]

        if adjust_pop:
            self.n_pop = 5 * self.n_dim * self.n_obj

            # If we are running under MPI, expand population to fully exploit all processors
            if self._running_under_mpi:
                self.n_pop = int(np.ceil(self.n_pop / self.comm.size) * self.comm.size)

            self.pop = np.concatenate(
                (
                    self.pop,
                    self.rng.uniform(
                        self.lb, self.ub, size=(self.n_pop - 1, self.n_dim)
                    ),
                )
            )
            self.f, self.cr = create_f_cr(
                self.adaptivity, self.f, self.cr, self.n_pop, self.rng
            )

            if self._running_under_mpi:
                self.pop, self.f, self.cr = self.comm.bcast(
                    (self.pop, self.f, self.cr), root=0
                )

            self.fit, self.con = self(self.pop)

        self.update()

        # Set generation counter to 0
        self.generation = 0

        # Mark class as initialized
        self._is_initialized = True

    @property
    def is_initialized(self):
        """bool: True if the algorithm has been initialized, False if not."""
        return self._is_initialized

    def __iter__(self):
        """
        This class is an iterator itself.

        Raises
        ------
        RuntimeError
            If this class is being used as an iterator before it has been initialized.
        """
        if not self._is_initialized:
            raise RuntimeError("NSDE is not yet initialized.")
        return self

    def __next__(self):
        """
        Main iteration.

        Returns
        -------
        NSDE
            The new state at the next generation.
        """
        if (
            self.generation < self.max_gen
            and self.dx > self.tolx
            and self.df > self.tolf
        ):
            # Create a new population and mutation/crossover parameters
            pop_new, f_new, cr_new = self.procreate()

            # Ensure all processors have the same updated population and mutation/crossover parameters
            if self._running_under_mpi:
                pop_new, f_new, cr_new = self.comm.bcast(
                    (pop_new, f_new, cr_new), root=0
                )

            # Evaluate the fitness of the new population
            fit_new, con_new = self(pop_new)

            # Update the class with the new data
            self.update(pop_new, fit_new, con_new, f_new, cr_new)

            # Compute spreads and update generation counter
            if self.n_obj == 1:
                self.dx = np.linalg.norm(self.pop[0] - self.pop[-1])
                self.df = np.abs(self.fit[0] - self.fit[-1])
            else:
                pareto = self.fit[self.fronts[0]]
                self.pareto_lb = np.minimum(self.pareto_lb, np.min(pareto, axis=0, keepdims=True))
                self.pareto_ub = np.maximum(self.pareto_ub, np.max(pareto, axis=0, keepdims=True))
                pareto_norm = 1 + (pareto - self.pareto_lb) / (self.pareto_ub - self.pareto_lb)
                self.hv = hv.hv(pareto_norm, 2.1 * np.ones(self.n_obj))
            self.generation += 1

            # Return the new state
            return self
        else:
            raise StopIteration

    def __call__(self, pop):
        """
        Evaluate the fitness of the given population.

        Parameters
        ----------
        pop : array_like
            List of chromosomes of the individuals in the population

        Returns
        -------
        fit : np.array
            Fitness of the individuals in the given population
        con : np.array or None
            Constraint violations of the individuals in the given population if present. None otherwise.

        Notes
        -----
        If this class has an MPI communicator the individuals will be evaluated in parallel.
        Otherwise function evaluation will be serial.
        """
        if self.is_initialized:
            fit = np.empty((self.n_pop, self.n_obj))
            con = None if self.n_con is None else np.empty((self.n_pop, self.n_con))
        else:
            fit = pop.shape[0] * [None]
            con = None

        def handle_result(_v, _i, _fit, _con):
            if isinstance(_v, tuple):
                _fit[_i] = np.asarray(_v[0])
                c = np.asarray(_v[1])
                if _con is None:
                    _con = np.empty((pop.shape[0], c.size))
                _con[_i] = c
            else:
                _fit[_i] = _v
            return _fit, _con

        # Evaluate generation
        if self._running_under_mpi:
            # Construct run cases
            cases = [((item, ii), None) for ii, item in enumerate(pop)]

            # Pad the cases with some dummy cases to make the cases divisible amongst the procs.
            extra = len(cases) % self.comm.size
            if extra > 0:
                for j in range(self.comm.size - extra):
                    cases.append(cases[-1])

            # Compute the fitness of all individuals in parallel using MPI
            results = concurrent_eval(
                self.fobj, cases, self.comm, allgather=True, model_mpi=self.model_mpi
            )

            # Gather the results
            for result in results:
                retval, err = result
                if err is not None or retval is None:
                    raise Exception(err)
                else:
                    fit, con = handle_result(*retval, fit, con)
        else:
            # Evaluate the population in serial
            for idx, ind in enumerate(pop):
                val = self.fobj(ind)
                fit, con = handle_result(val, idx, fit, con)

        # Turn all NaNs in the fitnesses into infs
        fit = np.reshape(np.where(np.isnan(fit), np.inf, fit), (pop.shape[0], -1))
        if con is not None:
            con = np.reshape(np.where(np.isnan(con), np.inf, con), (pop.shape[0], -1))
        return fit, con

    def run(self):
        for _ in self:
            pass

    def procreate(self):
        """
        Generate a new population using the selected evolution strategy.

        Returns
        -------
        pop_new : np.array
            Chromosomes of the individuals in the next generation
        f_new : np.array
            New set of mutation rates
        cr_new : np.array
            New set of crossover probabilities
        """
        pop_old_norm = (np.copy(self.pop) - self.lb) / self.range
        pop_new_norm = np.empty_like(pop_old_norm)

        # If there are constraints, augment the fitness to penalize infeasible individuals while procreating.
        # This stops the best and rand-to-best strategies to keep the best infeasible individual alive indefinitely.
        if self.n_con and False:
            fit = np.where(
                np.any(self.con >= 1e-6, axis=1, keepdims=True),
                np.linalg.norm(self.con, axis=1, keepdims=True) + np.max(self.fit),
                self.fit,
            )
        else:
            fit = self.fit

        if self.adaptivity == 0 or self.adaptivity == 1:
            if self.adaptivity == 0:
                # No adaptivity. Use static f and cr.
                f_new = self.f
                cr_new = self.cr
            else:
                # Simple adaptivity. Use new f and cr.
                f_new = np.where(
                    self.rng.uniform(size=self.n_pop) < 0.9,
                    self.f,
                    self.rng.uniform(size=self.n_pop) * 0.9 + 1,
                )
                cr_new = np.where(
                    self.rng.uniform(size=self.n_pop) < 0.9,
                    self.cr,
                    self.rng.uniform(size=self.n_pop),
                )

            for idx in range(self.n_pop):
                pop_new_norm[idx], _, _ = self.strategy(
                    idx, pop_old_norm, fit, self.fronts, f_new, cr_new, self.rng, False
                )
        else:
            # Complex adaptivity. Mutate f and cr.
            f_new = np.copy(self.f)
            cr_new = np.copy(self.cr)

            for idx in range(self.n_pop):
                pop_new_norm[idx], f_new[idx], cr_new[idx] = self.strategy(
                    idx, pop_old_norm, fit, self.fronts, self.f, self.cr, self.rng, True
                )

        pop_new = self.lb + self.range * np.asarray(pop_new_norm)
        return pop_new, f_new, cr_new

    def update(self, pop_new=None, fit_new=None, con_new=None, f_new=None, cr_new=None):
        """
        Update the population (and f/cr if self-adaptive).

        Parameters
        ----------
        pop_new : np.array or None, optional
            Proposed new population resulting from procreation
        fit_new : np.array or None, optional
            Fitness of the individuals in the new population
        con_new : np.array or None, optional
            Constraint violations of the individuals in the new population
        f_new : np.array or None, optional
            New set of mutation rates
        cr_new : np.array or None, optional
            New set of crossover probabilities

        Notes
        -----
        Individuals in the old population will only be replaced by the new ones if they have improved fitness.
        Mutation rate and crossover probabilities will only be replaced if self-adaptivity is turned on and if their
        corresponding individuals have improved fitness.
        """
        if self.n_obj == 1:
            self._update_single(pop_new, fit_new, con_new, f_new, cr_new)
        else:
            self._update_multi(pop_new, fit_new, con_new, f_new, cr_new)

    def _update_single(
        self, pop_new=None, fit_new=None, con_new=None, f_new=None, cr_new=None
    ):
        if self.n_con:
            cs = np.sum(
                np.where(np.greater(self.con, self.tolc), self.con, 0.0),
                axis=1
            )
        else:
            cs = 0

        if (
            pop_new is not None
            and fit_new is not None
            and f_new is not None
            and cr_new is not None
        ):
            if self.n_con:
                c_new = np.all(con_new <= self.tolc, axis=1)
                c_old = np.all(self.con <= self.tolc, axis=1)
                cs_new = np.sum(
                    np.where(np.greater(con_new, self.tolc), con_new, 0.0), axis=1
                )

                improved_indices = np.argwhere(
                    ((c_new & c_old) & (fit_new <= self.fit).flatten())
                    + (c_new & ~c_old)
                    + ((~c_new & ~c_old) & (cs_new <= cs))
                )

                self.con[improved_indices] = con_new[improved_indices]
                cs[improved_indices] = cs_new[improved_indices]
            else:
                improved_indices = np.argwhere((fit_new <= self.fit).flatten())

            self.pop[improved_indices] = pop_new[improved_indices]
            self.fit[improved_indices] = fit_new[improved_indices]

            if self.adaptivity != 0:
                self.f[improved_indices] = f_new[improved_indices]
                self.cr[improved_indices] = cr_new[improved_indices]

        # Sort population so the best individual is always the first
        idx_sort = np.argsort(
            self.fit.flatten() + np.where(cs != 0.0, cs * np.max(self.fit), 0.0)
        )

        self.pop = self.pop[idx_sort]
        self.fit = self.fit[idx_sort]
        if self.n_con:
            self.con = self.con[idx_sort]

        if self.adaptivity != 0:
            self.f = self.f[idx_sort]
            self.cr = self.cr[idx_sort]

    def _update_multi(
        self, pop_new=None, fit_new=None, con_new=None, f_new=None, cr_new=None
    ):
        if (
            pop_new is not None
            and fit_new is not None
            and f_new is not None
            and cr_new is not None
        ):
            self.pop = np.concatenate((self.pop, pop_new))
            self.fit = np.concatenate((self.fit, fit_new))
            if self.n_con:
                self.con = np.concatenate((self.con, con_new))

            if self.adaptivity != 0:
                self.f = np.concatenate((self.f, f_new))
                self.cr = np.concatenate((self.cr, cr_new))

        if self.n_con:
            fronts = sorting.nonDominatedSorting(self.fit, self.con, self.n_pop)
        else:
            fronts = sorting.nonDominatedSorting(self.fit, self.n_pop)
        fronts[-1] = np.asarray(fronts[-1])[
            sorting.crowdingDistanceSorting(self.fit[fronts[-1]])[
                : (self.n_pop - sum(len(f) for f in fronts[:-1]))
            ]
        ].tolist()
        new_idxs = []
        counter = 0
        self.fronts = []
        for front in fronts:
            new_idxs += front
            self.fronts += [list(range(counter, counter + len(front)))]
            counter += len(front)

        self.pop = self.pop[new_idxs]
        self.fit = self.fit[new_idxs]
        if self.n_con:
            self.con = self.con[new_idxs]

        if self.adaptivity != 0:
            self.f = self.f[new_idxs]
            self.cr = self.cr[new_idxs]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from os import path
import sys
import setuptools

here = path.abspath(path.dirname(__file__))

__version__ = re.findall(
    r"""__version__ = \"(.+)\"""",
    open(path.join(here, "nsde/__init__.py")).read(),
)[0]


with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


ext_modules = [
    Extension(
        "nsde.sorting",
        ["src/sorting/sorting_module.cpp", "src/sorting/sorting.cpp"],
        include_dirs=[
            "src/sorting/include",
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
        ],
        language="c++",
    ),
    Extension(
        "nsde.hv",
        ["src/hv/hv_module.cpp", "src/hv/hv.c"],
        include_dirs=[
            "src/hv/include",
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
        ],
        language="c++",
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """
    # pybind11 2.4.x seems to be incompatible with c++17
    # flags = ["-std=c++17", "-std=c++14", "-std=c++11"]
    flags = ["-std=c++14", "-std=c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError("Unsupported compiler -- at least C++11 support " "is needed!")


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {"msvc": ["/EHsc"], "unix": []}
    l_opts = {"msvc": [], "unix": []}

    if sys.platform == "darwin":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.7"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        opts.append("-O2")
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


setup(
    name="nsde",
    version=__version__,
    description="Non-dominated Sorting Differential Evolution (NSDE) Algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Daniel de Vries",
    author_email="danieldevries6@gmail.com",
    packages=["nsde"],
    ext_modules=ext_modules,
    python_requires=">=3.6, <4",
    install_requires=[
        "numpy>=1.17",
        "tqdm<5,>=4.32",
        "pybind11>=2.4",
    ],
    setup_requires=["pybind11>=2.4"],
    extras_require={"openmdao": "openmdao<2.10,>=2.8"},
    tests_require=["pytest", "parameterized"],
    cmdclass={"build_ext": BuildExt},
    url="https://github.com/daniel-de-vries/nsde",
    download_url="https://github.com/daniel-de-vries/nsde/archive/v{0}.tar.gz".format(
        __version__
    ),
    keywords=[
        "optimization",
        "black-box",
        "data science",
        "differential",
        "evolution",
        "evolutionary",
        "algorithms",
        "non-dominated",
        "crowding distance",
        "single-objective",
        "multi-objective",
    ],
    license="MIT License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: C++",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    zip_safe=False,
)
